#======================================================================
# ZMapper - use a delta Z map (matrix) or Barycentric interpolation
# to adjust Z movements for each extrusion move and insert the Z into 
# each line of code. This is a Z compensation to correct for minor
# mechanical and calibration issues on a delta printer. You must
# calibrate the delta radius, tower lengths and Z=0 before applying
# this script.
#
# Before: G1 X-4.24 Y29.08  E13.124
# After:  G1 X-4.24 Y29.08 Z34.4 E13.124
#
# The Z is the current layer height in mm plus the delta Z value at
# that X-Y position.
#
# USAGE: ZMapper.py -i gcodefile.gcode -n -t <trianglefile> -z heightmapfile.csv [-q]
#
# the -i file must have a ".gcode" extension
# the output file will be renamed as inptfilename-ZMapped.gcode
#
# only one of -t or -z is needed.
#
# -q prevents the postprocessor from running, this way you can disable
# the script without deleting the postprocessor line.
#
# Copyright (C) 2015 Michael Hackney
#======================================================================
import array
import sys
import os
import string
import csv
import getopt
import math

#---------------------------------------------------------------------------------------------
# Test if pointT is inside the triangle described by pointA, pointB and pointC using
# Barycentric coordinates. If it is, simply calculate the deltaZ for the test point
# using Barycentric interpolation of the pre-calculated Barycentric coordinates.
#
def deltaZfromTriangle(pointT, triangle):
    
    pointA = triangle[0]
    pointB = triangle[1]
    pointC = triangle[2]
    
    denomintor = (((pointB[1] - pointC[1]) * (pointA[0] - pointC[0])) + ((pointC[0] - pointB[0]) * (pointA[1] - pointC[1])))
                
    a = (((pointB[1] - pointC[1]) * (pointT[0] - pointC[0])) + ((pointC[0] - pointB[0]) * (pointT[1] - pointC[1]))) / denomintor
    b = (((pointC[1] - pointA[1]) * (pointT[0] - pointC[0])) + ((pointA[0] - pointC[0]) * (pointT[1] - pointC[1]))) / denomintor
    c = 1 - a - b

    if ((0 <= a) and (a <= 1) and (0 <= b) and (b <= 1) and (0 <= c) and (c <= 1)):
        return format((a * pointA[2]) + (b * pointB[2]) + (c * pointC[2]),'.2f')
    else:
        return "not found"        

#---------------------------------------------------------------------------------------------
# basez is the current slice height, x and y are the postions from the center of the bed (0,0)
#
def getZ(basez, x, y):
    X = x/10 + 14
    Y = y/10 + 14
    
    for row in heightMap:
        for e in row:
            print '%1.2f' % e,
        print "\n"
        
    print X
    print Y
    Z = basez + heightMap[X][Y]
    print Z

#---------------------------------------------------------------------------------------
# Determine which slicer create the gcode
#
def getSlicer(line):
    if line.startswith('; KISSlicer'):
        return "KISSlicer"
    elif line.startswith(';Generated with Cura_SteamEngine'):
        return "Cura"
    else:
        return "Unrecognized slicer"

#---------------------------------------------------------------------------------------
# Determine if the line is a begin layer section, if so, return the layer height (in mm)    
#
def find_begin_layer(line, infile, slicer):
    if slicer == "KISSlicer":
        if line.startswith('; BEGIN_LAYER_OBJECT z='):
            zIndex = line.find('z=')
            return line[zIndex+2:]
    elif slicer == "Cura":
        if line.startswith(';LAYER:'):
            layerIndex = line.find(':')
            layer = line[layerIndex+1:]
            for templine in infile:
                if templine.startswith('G0'):
                    parts = templine.split()

                    for part in parts:
                        firstChar = part[0]
                        
                        if firstChar == 'Z':
                            zIndex = float(part[1:])
                            return zIndex

    return False

#-----------------------------------------------------------------------------------------------------
# Determine if the line is a G0 or G1 move. If the line is a normal extrusion move, return the 
# line substrings, otherwise return False.
#
def find_extrude_segment(line, slicer):
    
    if slicer == "KISSlicer":
        # only process lines that start with a G1 movement
        if line.startswith('G1 X'):
            # make sure it isnt a prime or Z lift movement
            if 'Z' in line:
                return False
        lineParts = line.split()
        return lineParts
    
    elif slicer == "Cura":
        if line.startswith('G0') or line.startswith('G1'):
            lineParts = line.split()
            return lineParts

#----------------------------------------------------------------------------------------
# Use the X and Y positions in lineParts to lookup the Z height adjustment, add that to
# layerHeight and return a formatted gcode line. Currently supports two types of mapping:
#    1) a matrix of delta Z values
#    2) Barymetric Interpolation over a set of triangles   
#
def mapZ(lineParts, layerHeight, interpolationType):
    X = 0.0
    Y = 0.0
    Z = 0.0
    deltaZ = 0.0
    
    # perform Barycentric Interpolation
    if interpolationType == 1:

        # find X, Y for the move
        for part in lineParts:

            firstChar = part[0]
            
            if firstChar == 'X':
                X = float(part[1:])
            elif firstChar == 'Y':
                Y = float(part[1:])
            #elif firstChar == 'Z':
                #Z = float(part[1:])

        if ((X >= 0) and (Y >= 0)):
            for triangle in NEquad:
                deltaZ = deltaZfromTriangle([X,Y], triangle)
                if deltaZ != "not found":
                    break
    
        elif ((X >= 0) and (Y < 0)):
            for triangle in SEquad:
                deltaZ = deltaZfromTriangle([X,Y], triangle)
                if deltaZ != "not found":
                    break

        elif ((X < 0) and (Y < 0)):
            for triangle in SWquad:
                deltaZ = deltaZfromTriangle([X,Y], triangle)
                if deltaZ != "not found":
                    break
            
        elif ((X < 0) and (Y < 0)):
            for triangle in NWquad:
                deltaZ = deltaZfromTriangle([X,Y], triangle)
                if deltaZ != "not found":
                   break       

    # use matrix lookup
    else:
        # Here's where I'd like to do a linear interpolation to get a better Z estimate.
        #

        # find X, Y for the move and divide it by 10 for the 10mmx10mm cell size
        for part in lineParts:
            firstChar = part[0]
        
            if firstChar == 'X':
                X = part[1:]   
                X = int(round(float(X)/10) + columns/2)
            
            elif firstChar == 'Y':
                Y = part[1:]
                Y = int(round(float(Y)/10) + rows/2)

        deltaZ = heightMap[X][Y]
    
    if deltaZ == "not found":
        newZ = 0.0
    else:
        newZ = float(deltaZ) + float(layerHeight)

    zLine = 'Z' + str(newZ)

    lineParts.insert(3, zLine.strip())
    newCodeLine = string.join(lineParts, " ")

    return newCodeLine

#--------------------------------------------------------------------------
# Read the delta Z matrix from a CSV file and create a two dimensional list
#
def makeDeltaZ(heightmap):    
    with open(heightmap) as file:
        reader = csv.reader(file)
        htmap = list(reader)
        return htmap
                           
#====================================================
# __main__
#====================================================
if __name__ == '__main__':

    global NEquad
    global SEquad
    global SWquad
    global NWquad
    global heightMap

    try:
        opts, args = getopt.getopt(sys.argv[1:],"i:t:z:q:h")
    except getopt.GetoptError as err:
        print str(err)
        sys.exit('usage: ZMapper.py -i <gcodefile> -t <trianglefile> -z <heightmapfile> [-q] [-h]')
    
    interpolationType = 0
    
    for o, a in opts:
        if o == "-q":
            sys.exit()
        elif o == "-i":
            gcodefn = a
        elif o == "-t":
            trianglefn = a
            interpolationType = 1
        elif o == "-z":
            heightfn = a
            interpolationType = 2
        elif o == "-h":
            print "ZMapper adjusts the Z position for every point in the passed-in gcode file. This Z position"
            print "can be calculated in one of two ways. The first uses Barycentric Interpolation, the -t option."
            print "The second uses a matrix to lookup the delta Z adjustment, the -z option. You must specify"
            print "one either -t or -z with the path to the appropriate file."
    
    if interpolationType == 1:
        execfile(trianglefn)
                
    elif interpolationType == 2:
        heightMap = makeDeltaZ(str(heightfn))
        columns = len(heightMap)
        rows = len(heightMap[:])
    
    else:
        print "you must specify either a triangle or heightmap file"
    
    # Now let's map the Zs
    #
    infile = open(gcodefn, 'rb')
    currentZ = 0
    
    # Create a new file to write the Z mapped gcode to, we will not harm the original gcode!
    # 
    newfn = string.replace(gcodefn, ".gcode","-ZMapped.gcode")
    with open(newfn, 'w') as outfile:
        outfile.write(';---------------------------------------------------------------')
        outfile.write("\n")    
        outfile.write('; Processed with ZMapper using ')
        if interpolationType == 1:
            outfile.write('Barycentric Interpolation')
        else:
            outfile.write('Height Map Matrix')
        
        outfile.write("\n")
        outfile.write(';---------------------------------------------------------------')
        outfile.write("\n")
        
        # This is where the work is done to parse the gcode and look  
        # for "G1" moves that need to have an explicit Z value added.
        #
        
        # Determine which slicer was used
        slicer = getSlicer(infile.readline())

        infile.seek(0,0)
        
        for line in infile:
            
            layerHeight = find_begin_layer(line, infile, slicer)

            if layerHeight != False:
                currentZ = layerHeight

            lineParts = find_extrude_segment(line, slicer)
            if lineParts != False:
                if lineParts != None:
                    mappedLine = mapZ(lineParts, currentZ, interpolationType)
                    outfile.write(mappedLine)
                    outfile.write("\n")
                else:
                    outfile.write(line.replace('\r',''))
            
        infile.close()
        outfile.close()